/*
 *************************************************************************
 *
 * serdisp_connect_usb.c
 * routines for accessing usb devices
 *
 *************************************************************************
 *
 * copyright (C) 2003-2006  wolfgang astleitner
 * email     mrwastl@users.sourceforge.net
 *
 *************************************************************************
 * This program is free software; you can redistribute it and/or modify   
 * it under the terms of the GNU General Public License as published by   
 * the Free Software Foundation; either version 2 of the License, or (at  
 * your option) any later version.                                        
 *                                                                        
 * This program is distributed in the hope that it will be useful, but    
 * WITHOUT ANY WARRANTY; without even the implied warranty of             
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      
 * General Public License for more details.                               
 *                                                                        
 * You should have received a copy of the GNU General Public License      
 * along with this program; if not, write to the Free Software            
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA              
 * 02111-1307, USA.  Or, point your browser to                            
 * http://www.gnu.org/copyleft/gpl.html                                   
 *************************************************************************
 */

/*#define DEBUG_SIGNALS 1*/
/*#define ENABLE_FTDI 1*/

/* use usb_clear_halt() instead of usb_bulk_read() (which seems to cause problems on some linux versions) */
#define IOW24_USE_USB_CLEAR_HALT 1

#include "../config.h"

#include <syslog.h>

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
/*#include <fcntl.h>*/
#include <string.h>
#include <strings.h>
#include <sys/ioctl.h>

/*#include <sys/resource.h>*/

#include "serdisplib/serdisp_connect.h"
#include "serdisplib/serdisp_connect_usb.h"
#include "serdisplib/serdisp_tools.h"
#include "serdisplib/serdisp_messages.h"



/* supported usb-devices */
typedef struct serdisp_usbdevice_s {
  int   vendorID;
  int   productID;
  int   bcdDevice;
  short deviceID;
  int   streamsize;
  byte  maxIO;
} serdisp_usbdevices_t;


/* parallel port: signal names + aliases. needed for customisable wiring */
serdisp_usbdevices_t usbdevices[] = {
/*  vendorID  productID   bcdDevice    deviceID             streamsize         maxIO */
   {0x07C0,    0x1501,        -1,      SDHWT_USBIOW24,      8,                 16 }
#ifdef ENABLE_FTDI   
  ,{0x0403,    0x6010,    0x0500,      SDHWT_USBFTDI2232,   SD_USB_BUFFERSIZE, 16 }
  ,{0x0403,    0x6001,    0x0200,      SDHWT_USBFTDI245,    SD_USB_BUFFERSIZE,  8 }
  ,{0x0403,    0x6001,    0x0400,      SDHWT_USBFTDI245,    SD_USB_BUFFERSIZE,  8 }
#endif
};



/* local impl. of usb_detach_kernel_driver_np */
#ifndef HAVE_USB_DETACH_KERNEL_DRIVER_NP

int usb_detach_kernel_driver_np(usb_dev_handle* dev, int interface) {
  sd_debug(0,  "usb_detach_kernel_driver_np(): *** this function is not provided in this version of libusb.");
  sd_debug(0,  "usb_detach_kernel_driver_np(): ***   either upgrade libusb to at least version 0.8 or you'll");
  sd_debug(0,  "usb_detach_kernel_driver_np(): ***   have to detach kernel modules by yourself.");
  return 0;
}
#endif /*HAVE_USB_DETACH_KERNEL_DRIVER_NP */



/* *********************************
   serdisp_CONN_t* SDCONNusb_open(sdcdev)
   *********************************
   opens an usb device for a serdisp lcd
   *********************************
   sdcdev   ... device identifier. format:
   
        devid = 'USB:' [ occurrence ':' ] ( PRODUCT | DEVICENAME )
      PRODUCT = vendorID '/' productID [ '/' bcdDevice ]
   DEVICENAME = manufacturer '/' productdesc

               examples:
                "USB:0x0403/0x6010/0x0500"   (vendorID = 0x0403, productID = 0x6010, bcdDevice = 0x500)
                "USB:403/6010"               (the same, w/o bcdDevice (first device matching 403/6010 is returned))
                "USB:FTDI/DLP2232M"          (identified through manufacturer info and product description)
                "USB:0:FTDI/DLP2232M"        (the same, fist occurrence (default))
                "USB:1:FTDI/DLP2232M"        (the same, but 2nd occurrence (if > 1 DLP2232M are connected))

   *********************************
   returns a serdisp connect descriptor or (serdisp_CONN_t*)0 if operation was unsuccessful

*/
serdisp_CONN_t* SDCONNusb_open(const char sdcdev[]) {
  serdisp_CONN_t* sdcd = 0;
  struct usb_bus* bus = 0;
  struct usb_device* dev = 0;
  usb_dev_handle* usb_dev = 0;
  char* devname;
  char protocol[20];
  char buffer[10];
  char* idx;  /* index of ':' in protocol:devicename */
  char* endptr;  /* for strtol */
  int occurrence = 0;
  int occ_count = 0;
  int vendorID = -1;
  int productID = -1;
  int bcdDevice = -1;

  int found = 0;          /* usb device found in device list */
  int usbdev_found = 0;   /* usb device found in usbdevices[] */
  int i = 0;

  serdisp_usbdev_t* usbitems = (serdisp_usbdev_t*)0;

  devname = (char*)sdcdev;
  idx = index(devname, ':');

  if (serdisp_ptrdistance(idx, devname) >= 18) {
    sd_error(SERDISP_ENXIO, "SDCONNusb_open(): invalid protocol (too long)");
    return (serdisp_CONN_t*)0;
  }

  sdtools_strncpy(protocol, devname, serdisp_ptrdistance(idx, devname));
  devname = (idx+1);

  idx = index(devname, ':');

  if (idx) {
    if (serdisp_ptrdistance(idx, devname) >= 8) {
      sd_error(SERDISP_ENXIO, "SDCONNusb_open(): invalid occurence ID");
      return (serdisp_CONN_t*)0;
    }
    sdtools_strncpy(buffer, devname, serdisp_ptrdistance(idx, devname));
    devname = (idx+1);
    occurrence = (int)strtol(buffer, 0, 10);
  }

  idx = index(devname, '/');
  if (idx && serdisp_ptrdistance(idx, devname) < 7) {
    sdtools_strncpy(buffer, devname, serdisp_ptrdistance(idx, devname));
    vendorID = (int)strtol(buffer, &endptr, 16);
    if (buffer == endptr || (*endptr != '\0') )  /* value invalid */
      vendorID = -1;
    else
      devname = (idx+1);  
  }

  
  if (vendorID != -1) {
    idx = index(devname, '/');
    if (!idx) idx = devname + strlen(devname);
    if (serdisp_ptrdistance(idx, devname) < 7) {
      sdtools_strncpy(buffer, devname, serdisp_ptrdistance(idx, devname));
      productID = (int)strtol(buffer, &endptr, 16);
      if (buffer == endptr || (*endptr != '\0') ) { /* value invalid */
        productID = -1;
        vendorID = -1;
      } else
        devname = idx;
    }
  }

  if (productID != -1 && devname[0] == '/') {
    devname++;
    idx = devname + strlen(devname);
    
    if (serdisp_ptrdistance(idx, devname) < 7) {
      sdtools_strncpy(buffer, devname, serdisp_ptrdistance(idx, devname));
      bcdDevice = (int)strtol(buffer, &endptr, 16);
      if (buffer == endptr || (*endptr != '\0') ) { /* value invalid */
        productID = -1;
        vendorID = -1;
        bcdDevice = -1;
      }
    }
  }

  usb_init();
  if (usb_find_busses() < 0) {
    sd_error(SERDISP_ENXIO, "SDCONNusb_open(): usb_find_busses() failed. error: %s", strerror(errno));
    return (serdisp_CONN_t*)0;
  }
  if (usb_find_devices() < 0) {
    sd_error(SERDISP_ENXIO, "SDCONNusb_open(): usb_find_devices() failed. error: %s", strerror(errno));
    return (serdisp_CONN_t*)0;
  }

  /* scan usb bus for device given by sdcdev */
  bus = usb_get_busses();
  while (!found && bus) {
    dev = bus->devices;
    while (!found && dev) {
      usb_dev = usb_open(dev);
      if (usb_dev) {
        if ( (dev->descriptor.idVendor ==  vendorID) && 
             (dev->descriptor.idProduct == productID) &&
             (bcdDevice == -1 || dev->descriptor.bcdDevice == bcdDevice)
           ) {
          if (occurrence == occ_count) {
            found = 1;
            sd_debug(1, "SDCONNusb_open(): device found: 0x%04x/0x%04x/0x%04x", 
                         dev->descriptor.idVendor, dev->descriptor.idProduct, dev->descriptor.bcdDevice);
            i = 0;        
            while (!usbdev_found && i < sizeof(usbdevices) / sizeof(serdisp_usbdevices_t)) {
              if ( (vendorID == usbdevices[i].vendorID) && 
                   (productID == usbdevices[i].productID) &&
                   (bcdDevice == -1 || usbdevices[i].bcdDevice == -1 || bcdDevice == usbdevices[i].bcdDevice)
                 ) {
                 usbdev_found = 1;
              } else {
                i++;
              }
            }

            if (!usbdev_found) {
              sd_error(SERDISP_ENOTSUP, "SDCONNusb_open(): device 0x%04x/0x%04x not found in supported devices list", 
                                         dev->descriptor.idVendor, dev->descriptor.idProduct);
            }

          } else
            occ_count++;  /* device would be ok, but we search for another one */
        }
      }
      if (!found) usb_close(usb_dev);
     if (!found) dev = dev->next;
    }
    bus = bus->next;
  }

  if (found && usbdev_found) {
    int rc;
    
    if (! (sdcd = (serdisp_CONN_t*)sdtools_malloc(sizeof(serdisp_CONN_t)) ) ) {
      sd_error(SERDISP_EMALLOC, "SDCONNusb_open(): unable to allocate memory for sdcd");
      return (serdisp_CONN_t*)0;
    }
    memset(sdcd, 0, sizeof(serdisp_CONN_t));


    sdcd->sdcdev = (char*)sdcdev;

    sdcd->conntype = SERDISPCONNTYPE_PARPORT;
    sdcd->hardwaretype = SDHWT_USB;

    if (! (sdcd->extra = (void*)sdtools_malloc(sizeof(serdisp_usbdev_t)) ) ) {
      sd_error(SERDISP_EMALLOC, "SDCONNusb_open(): unable to allocate memory for sdcd->extra");
      free(sdcd);
      return (serdisp_CONN_t*)0;
    }
    memset(sdcd->extra, 0, sizeof(serdisp_usbdev_t));

    usbitems = (serdisp_usbdev_t*)(sdcd->extra);
    
    /* some initialisations */
    usbitems->devID = i;  /* index in usbdevices[] */
    usbitems->streamsize = usbdevices[usbitems->devID].streamsize;
    
    if (! (usbitems->stream = (char*) sdtools_malloc( sizeof(char) * usbitems->streamsize ) ) ) {
      sd_error(SERDISP_EMALLOC, "sDCONNusb_open(): cannot allocate stream buffer");
      free(sdcd->extra);
      free(sdcd);
      return (serdisp_CONN_t*)0;
    }

    usbitems->streampos = 0;

    usbitems->dev = dev;
    usbitems->usb_dev = usb_dev;
    usbitems->read_timeout = 5000;
    usbitems->write_timeout = 5000;


    /* claim interfaces */
    for (i = 0; i < dev->config->bNumInterfaces; i++) {
      /* detach from kernel module if bound to one */
      usb_detach_kernel_driver_np(usb_dev, i);

      if ((rc = usb_claim_interface(usb_dev, i)) < 0) {
        sd_error(SERDISP_EACCES, 
                 "SDCONNusb_open(): usb_claim_interface() unsuccessful for interface %d. rc=%d, error: %s", 
                 i, rc, strerror(errno)
                );
        free(((serdisp_usbdev_t*)(sdcd->extra))->stream);
        free(sdcd->extra);
        free(sdcd);
        return (serdisp_CONN_t*)0;
      } else {
        usbitems->claimed |= (1<<i);
        sd_debug(1, "SDCONNusb_open(): usb_claim_interface() successful for interface %d\n", i);
      }      
    }


    switch (usbdevices[usbitems->devID].deviceID) {
#ifdef ENABLE_FTDI   
      case SDHWT_USBFTDI2232:
      case SDHWT_USBFTDI245:
        if (usb_control_msg(usb_dev, 0x40, 0, 0, 1/*ftdi->index*/, NULL, 0, usbitems->write_timeout) != 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): resetting FTDI-chip failed");
        }

        if (usb_control_msg(usb_dev, 0x40, 0x0B, 0x0100 | 0xFF, 1/*ftdi->index*/, NULL, 0, usbitems->write_timeout) != 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): entering bitbang mode failed");
        }

        if ((rc = usb_control_msg(usb_dev, 0x40, 3, 1, 1/*ftdi->index*/, NULL, 0, usbitems->write_timeout)) != 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): setting baudrate failed");
        }
      break;
#endif      
      case SDHWT_USBIOW24: {
        /* only 1 altsetting and 1 endpoint */
        usbitems->in_ep = dev->config->interface[1].altsetting[0].endpoint[0].bEndpointAddress;  
      
        /* iow24: enable lcd */
        IOW_FILLSTREAM(usbitems->stream, IOW_LCD_ENABLE_REPORT, 0x01, 0,0,0,0,0,0);  
        if (usb_control_msg(
             usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
                           ) < 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): IOW/LCD enabling  failed");
        }

        /* iow24: enable i2c */
        IOW_FILLSTREAM(usbitems->stream, IOW_I2C_ENABLE_REPORT, 0x01, 0,0,0,0,0,0);  
        if (usb_control_msg(
              usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
                           ) < 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): IOW/I2C enabling  failed");
        }

        /* iow24: init i2c */
        usbitems->store = 0xCF | 0x10;  /* all leds off (0xCF, leds are active low), enable 1st cs-line (0x10) */
        IOW_FILLSTREAM(usbitems->stream, IOW_I2C_WRITE_REPORT, 0xC2, 0x70, (byte)(usbitems->store) ,0,0,0,0);
        if (usb_control_msg(
              usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
            ) < 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): IOW/I2C initial setting failed");
				}
        /* read acknowledge */
#ifdef IOW24_USE_USB_CLEAR_HALT
        if (usb_clear_halt(usbitems->usb_dev, usbitems->in_ep) < 0) {
#else        
        if (usb_bulk_read(usbitems->usb_dev, usbitems->in_ep, usbitems->stream, 8, usbitems->write_timeout) < 0) {
#endif        
            sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): IOW/I2c reading ack report failed");
        }

        usbitems->laststatus = 1;  /* set laststatus to COMMAND (==1) */
        usbitems->streampos = 0;
      }
      break;
      default:
        sd_error(SERDISP_ERUNTIME, "SDCONNusb_open(): switch(deviceID): hw type unsupported!!!");        
    }
  } else {
    sd_error(SERDISP_EACCES, "SDCONNusb_open(): unable to find or open a matching device");
    return (serdisp_CONN_t*)0;
  }

  return sdcd;
}



/* *********************************
   void SDCONNusb_close(sdcd)
   *********************************
   close the device occupied by serdisp
   *********************************
   sdcd     ... serdisp connect descriptor
   *********************************
*/
void SDCONNusb_close(serdisp_CONN_t* sdcd) {
  int i;
  serdisp_usbdev_t* usbitems = (serdisp_usbdev_t*)(sdcd->extra);

  switch (usbdevices[usbitems->devID].deviceID) {
#ifdef ENABLE_FTDI   
    case SDHWT_USBFTDI2232:
    case SDHWT_USBFTDI245:
    break;
#endif    
    case SDHWT_USBIOW24: {
    
      /* iow24: disable lcd */
      IOW_FILLSTREAM(usbitems->stream, IOW_LCD_ENABLE_REPORT, 0x00, 0,0,0,0,0,0);
      if (usb_control_msg(
            usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
                         ) <0) {
        sd_error(SERDISP_ERUNTIME, "SDCONNusb_close(): IOW/LCD shutdown failed");
      }

      /* iow24: de-init i2c signals */
      IOW_FILLSTREAM(usbitems->stream, IOW_I2C_WRITE_REPORT, 0xC2, 0x70,0xFF, 0,0,0,0);
      if (usb_control_msg(
            usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
                         ) < 0) {
        sd_error(SERDISP_ERUNTIME, "SDCONNusb_close(): IOW/I2C de-initialising failed");
			}

      /* iow24: disable i2c */
      IOW_FILLSTREAM(usbitems->stream, IOW_I2C_ENABLE_REPORT, 0x00, 0,0,0,0,0,0);  
      if (usb_control_msg(
            usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
                         ) <0) {
        sd_error(SERDISP_ERUNTIME, "SDCONNusb_close(): IOW/I2C shutdown failed");
      }
      /* i2c read acknowledge */

#ifdef IOW24_USE_USB_CLEAR_HALT
      if (usb_clear_halt(usbitems->usb_dev, usbitems->in_ep) < 0) {
#else        
      if (usb_bulk_read(
            usbitems->usb_dev, usbitems->in_ep, usbitems->stream, 8, usbitems->write_timeout
                       ) < 0) {
#endif
        sd_error(SERDISP_ERUNTIME, "SDCONNusb_close(): IOW/I2c reading ack report failed");
      }
    }
    break;
    default: /* should never ever occur */
      sd_error(SERDISP_ERUNTIME, "SDCONNusb_close(): switch-case: hwtype not defined!!!");        
  }


  /* unclaim interfaces */
  for (i = 0; i < usbitems->dev->config->bNumInterfaces; i++) {
    if (usbitems->claimed & (1<<i) ) {
      usb_release_interface(usbitems->usb_dev, i);
      usbitems->claimed ^= (1<<i);
    }
  }
  usb_close(usbitems->usb_dev);
  free(usbitems->stream);
  free(sdcd->extra);
}



/* *********************************
   void SDCONNusb_write(sdcd, ldata, flags)
   *********************************
   write a byte to the serdisp connect device
   *********************************
   sdcd   ... serdisp connect descriptor
   ldata  ... data to be written 
   flags  ... which bytes are to read
*/
void SDCONNusb_write(serdisp_CONN_t* sdcd, long ldata, byte flags) {
  SDCONNusb_writedelay(sdcd, ldata, flags, 0);
}

/* *********************************
   void SDCONNusb_writedelay(sdcd, ldata, flags, ns)
   *********************************
   write a byte to the serdisp connect device
   *********************************
   sdcd   ... serdisp connect descriptor
   ldata  ... data to be written 
   flags  ... which bytes are to read  (ignored in SDCONNusb_*)
   ns     ... nanoseconds delay after write
*/
void SDCONNusb_writedelay(serdisp_CONN_t* sdcd, long ldata, byte flags, long ns) {
  byte t_data = 0, t_flags = 0;
  serdisp_usbdev_t* usbitems = (serdisp_usbdev_t*)(sdcd->extra);

#ifdef DEBUG_SIGNALS
  {
    int l;
    int ddd = ((int) ((ldata & 0x000F0000) >> 8)) | ((int) (ldata & 0x000000FF) );
    for (l = 11; l >= 0; l--) {
      fprintf(stderr, "%s", ((ddd & (1 << l)) ? " #  " : "|   "  ));
      if (l % 4 == 0)
        fprintf(stderr, "  ");
    }
    fprintf(stderr, "\n");
  }
#endif    

  /* control byte */
  /* write bits 0-3 from input data and bits 4-7 from stored control byte */
  t_flags = (byte) ((ldata & 0x000F0000) >> 16);


  /* data byte */
  t_data = (byte) (ldata & 0x000000FF);

  switch (usbdevices[usbitems->devID].deviceID) {
#ifdef ENABLE_FTDI   
    case SDHWT_USBFTDI2232:
    case SDHWT_USBFTDI245:
      usbitems->stream[usbitems->streampos++] = t_data;

      if (ns|| 1==1)
         usbitems->stream[usbitems->streampos++] = t_data;

      if (usbitems->streampos+1 >= usbitems->streamsize)
        SDCONNusb_commit(sdcd);
    break;
#endif
    case SDHWT_USBIOW24: 
      {
        byte oldcs = (usbitems->store & 0x20) ? 1: 0;
        
        /* t_flags:  0000 00000
                              ^--- 0.. t_data == data byte, 1 .. t_data == command byte
         */

        if ((t_flags & 0x02) && (t_data != oldcs)) {  /* cs-line change */

          SDCONNusb_commit(sdcd);      

          usbitems->store &= 0xCF;
          usbitems->store |= ((t_data) ? 0x20 : 0x10);

          /* iow24: i2c: switch lines */
          IOW_FILLSTREAM(usbitems->stream, IOW_I2C_WRITE_REPORT, 0xC2, 0x70, (byte)(usbitems->store) ,0,0,0,0);

          if (usb_control_msg(
                usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
                             ) < 0) {
            sd_error(SERDISP_ERUNTIME, "SDCONNusb_writedelay(): IOW/I2C CS-line change failed");

#ifdef IOW24_USE_USB_CLEAR_HALT
          } else if (usb_clear_halt(usbitems->usb_dev, usbitems->in_ep) < 0) {
#else        
          } else if (usb_bulk_read(usbitems->usb_dev, usbitems->in_ep, usbitems->stream, 8, usbitems->write_timeout) < 0) {
#endif        
            /* read acknowledge */
            sd_error(SERDISP_ERUNTIME, "SDCONNusb_writedelay(): IOW/I2c reading ack report failed");
          }
        } else  if (usbitems->laststatus != (t_flags & 0x01) ) {
          SDCONNusb_commit(sdcd);
          usbitems->laststatus = (t_flags & 0x01);
        }

        if (t_flags <= 0x01) {
          usbitems->stream[2 + usbitems->streampos++] = t_data; /* byte 0 and 1 are needed for repID and value */
          if (usbitems->streampos+1 >= 6)
            SDCONNusb_commit(sdcd);      
        }
      }
      break;
  };

  sdcd->debug_count++;
}

/* *********************************
   byte SDCONNusb_read(sdcd, flags)
   *********************************
   read a byte from the serdisp connect device
   *********************************
   sdcd   ... serdisp connect descriptor
   flags  ... which bytes are to read
   *********************************
   returns a data read from the serdisp connect device
   =================================================

*/
long SDCONNusb_read(serdisp_CONN_t* sdcd, byte flags) {
  /*long ldata = 0;*/

  if (sd_runtime_error())
    return 0;

  return 0;
}



/* *********************************
   void SDCONNusb_commit(sdcd)
   *********************************
   commits an usb write stream
   *********************************
   sdcd   ... serdisp connect descriptor
*/
void SDCONNusb_commit(serdisp_CONN_t* sdcd) {
  serdisp_usbdev_t* usbitems = (serdisp_usbdev_t*)(sdcd->extra);

  if (usbitems->streampos > 0) {
    switch (usbdevices[usbitems->devID].deviceID) {
#ifdef ENABLE_FTDI   
      case SDHWT_USBFTDI2232:
      case SDHWT_USBFTDI245:
        if(usb_bulk_write(
            usbitems->usb_dev, 
            0x02/*ftdi->in_ep*/, 
            usbitems->stream, usbitems->streampos-1,
            usbitems->write_timeout
                          ) < 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_commit(): FTDI: usb_bulk_write() failed (cause: %s)", strerror(errno));
          sd_runtimeerror = 1;
        }
      break;
#endif
      case SDHWT_USBIOW24: {
        usbitems->stream[0] = IOW_LCD_WRITE_REPORT;  /* iow24: lcd reportID */
        usbitems->stream[1] = usbitems->streampos + ((usbitems->laststatus) ? 0 : 0x80);
        
        if (usb_control_msg(
              usbitems->usb_dev, USB_DT_HID, IOW_REQ_SET_REPORT, 0x200, 1, usbitems->stream, 8, usbitems->write_timeout
           ) < 0) {
          sd_error(SERDISP_ERUNTIME, "SDCONNusb_commit(): IOW/LCD write failed (cause: %s)", strerror(errno));
          sd_runtimeerror = 1;
        }
      }         
      break;
    }
    usbitems->streampos = 0;    
  }
}


