/*
 *************************************************************************
 *
 * serdisp_specific_sed133x.c
 * routines for controlling displays controlled by sed1330/sed1335
 *
 *************************************************************************
 *
 * copyright (C) 2003-2007  wolfgang astleitner
 * email     mrwastl@users.sourceforge.net
 *
 *************************************************************************
 * This program is free software; you can redistribute it and/or modify   
 * it under the terms of the GNU General Public License as published by   
 * the Free Software Foundation; either version 2 of the License, or (at  
 * your option) any later version.                                        
 *                                                                        
 * This program is distributed in the hope that it will be useful, but    
 * WITHOUT ANY WARRANTY; without even the implied warranty of             
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      
 * General Public License for more details.                               
 *                                                                        
 * You should have received a copy of the GNU General Public License      
 * along with this program; if not, write to the Free Software            
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA              
 * 02111-1307, USA.  Or, point your browser to                            
 * http://www.gnu.org/copyleft/gpl.html                                   
 *************************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>

#include "serdisplib/serdisp_connect.h"
/*#include "serdisplib/serdisp_specific_sed133x.h"*/
#include "serdisplib/serdisp_tools.h"
#include "serdisplib/serdisp_messages.h"
#if (SERDISP_VERSION_CODE > SERDISP_VERSION(1,97))
 #include "serdisplib/serdisp_colour.h"
#endif


/*#define OPT_USEOLDUPDATEALGO */

/*
 * command constants
 */

#define CMD_SYSTEMSET     0x40
#define CMD_SLEEPIN       0x53
#define CMD_DISPON        0x59
#define CMD_DISPOFF       0x58
#define CMD_SCROLL        0x44
#define CMD_CSRFORM       0x5D
#define CMD_CGRAMADR      0x5C
#define CMD_CSRDIRRIGHT   0x4C
#define CMD_CSRDIRLEFT    0x4D
#define CMD_CSRDIRUP      0x4E
#define CMD_CSRDIRDOWN    0x4F
#define CMD_HDOTSCR       0x5A
#define CMD_OVLAY         0x5B
#define CMD_CSRW          0x46
#define CMD_MWRITE        0x42

#define OPT_SAD              0 /* start address for graphics layer */
#define OPT_M0               0 /* 0: internal CG ROM             1: external CG ROM               */
#define OPT_M1               0 /* 0: CG RAM1/32 char             1: CG RAM1+RAM2/64 char          */
#define OPT_M2               0 /* 0: 8-pixel                     1: 16-pixel character height     */
#define OPT_WS               0 /* 0: single-panel                1: dual-panel drive              */
#define OPT_IV               1 /* 0: screen top-line correction  1: no screen top-line correction */
#define OPT_TL               0 /* 0: LCD-mode                    1: TV-mode                       */
#define OPT_DR               0 /* 0: normal operation            1: additional shift-clock cycles */
#define OPT_WF               1 /* 0: 16-line AC drive            1: two-frame AC-drive            */
#define OPT_FX               8 /* hor. character size  */
#define OPT_FY               8 /* vert. character size */

#define OPT_MX0              0
#define OPT_MX1              0 /* MX1/MX0 == 0/0  -> composition method = OR */
#define OPT_DM1              1
#define OPT_DM2              1 /* DM1/2 == 1/1    -> graphics mode */
#define OPT_OV               0 /* two-layer composition (needed for graphics mode) */

#define OPT_FC0              0
#define OPT_FC1              0 /* FC0/FC1 == 0/0  -> no cursor */
#define OPT_FP            0x01 /* FP = 000001 => 1st screen on, no flashing */

/*
 * values used mainly for initialization
 */


#define INTERFACE_8080      0
#define INTERFACE_6800      1


/* define shorts for signals to make coding simpler. must match order of serdisp_sed133x_wiresignals[] */
#define SIG_D0          (dd->sdcd->signals[ 0])
#define SIG_D1          (dd->sdcd->signals[ 1])
#define SIG_D2          (dd->sdcd->signals[ 2])
#define SIG_D3          (dd->sdcd->signals[ 3])
#define SIG_D4          (dd->sdcd->signals[ 4])
#define SIG_D5          (dd->sdcd->signals[ 5])
#define SIG_D6          (dd->sdcd->signals[ 6])
#define SIG_D7          (dd->sdcd->signals[ 7])
#define SIG_CS          (dd->sdcd->signals[ 8])
#define SIG_A0          (dd->sdcd->signals[ 9])
#define SIG_WR          (dd->sdcd->signals[10])
#define SIG_RD          (dd->sdcd->signals[11])
#define SIG_RESET       (dd->sdcd->signals[12])
#define SIG_BACKLIGHT   (dd->sdcd->signals[13])


/* different display types/models supported by this driver */
#define DISPID_SED133X  1


serdisp_wiresignal_t serdisp_sed133x_wiresignals[] = {
 /*  type   signame   actlow   cord  index */
   {SDCT_PP, "CS",         1,   'C',    8 }
  ,{SDCT_PP, "A0",         0,   'C',    9 }
  ,{SDCT_PP, "WR",         1,   'C',   10 }
  ,{SDCT_PP, "RD",         1,   'C',   11 }
  ,{SDCT_PP, "RESET",      1,   'D',   12 }
  ,{SDCT_PP, "BACKLIGHT",  0,   'D',   13 }
};

/* wirings are taken from graphlcd-base-0.1.2-pre4, docs/DRIVER.sed1330 */

serdisp_wiredef_t serdisp_sed133x_wiredefs[] = {
   {  0, SDCT_PP, "Original", "DATA8,CS:nSELIN,A0:INIT,WR:nAUTO,RD:nSTRB", "Original wiring"}
  ,{  1, SDCT_PP, "PowerLCD", "DATA8,CS:nAUTO,A0:INIT,WR:nSTRB,RD:nSELIN", "PowerLCD wiring"}
  ,{  2, SDCT_PP, "LCDProc",  "DATA8,CS:nSTRB,A0:nSELIN,WR:nAUTO,RD:INIT", "LCDProc wiring"}
  ,{  3, SDCT_PP, "Tweakers", "DATA8,CS:nSTRB,A0:nSELIN,WR:INIT,RD:nAUTO", "Tweakers wiring"}
  ,{  4, SDCT_PP, "YASEDW",   "DATA8,CS:nSELIN,A0:nAUTO,WR:nSTRB,RD:INIT", "YASEDW wiring"}
};

serdisp_options_t serdisp_sed133x_options[] = {
   /*  name          aliasnames   min    max mod int defines  */
   {  "WIDTH",       "",            0,   640,  8, 0,  ""}
  ,{  "HEIGHT",      "",            0,   256,  8, 0,  ""}
  ,{  "DELAY",       "",            0,    -1,  1, 1,  ""}
  ,{  "OSCILLATOR",  "OSC,FOSC", 1000, 10000,  1, 1,  ""}                /* kHz */
  ,{  "INTERFACE",   "MODE",        0,     1,  1, 0,  "8080=0,6800=1"}
};


/* internal typedefs and functions */

static void serdisp_sed133x_init      (serdisp_t*);
static void serdisp_sed133x_update    (serdisp_t*);
static int  serdisp_sed133x_setoption (serdisp_t*, const char*, long);
static void serdisp_sed133x_close     (serdisp_t*);

static void serdisp_sed133x_setpixel  (serdisp_t*, int, int, long);
static long serdisp_sed133x_getpixel  (serdisp_t*, int, int);



typedef struct serdisp_sed133x_specific_s {
  int  interfacemode;
  int  fosc;
} serdisp_sed133x_specific_t;


serdisp_sed133x_specific_t* serdisp_sed133x_internal_getStruct(serdisp_t* dd) {
  return (serdisp_sed133x_specific_t*)(dd->specific_data);
}


void serdisp_sed133x_transfer(serdisp_t* dd, int iscmd, byte item) {
  long item_split = 0;
  long td_clk1 = 0;
  long td_clk2 = 0;
  long td_clk3 = 0;
  int i;

  /* active-low signals are internally seen active-high because they will be auto-inverted later if needed */
  if (serdisp_sed133x_internal_getStruct(dd)->interfacemode == INTERFACE_6800) {
    td_clk1 = SIG_WR | SIG_RD;
    td_clk2 = SIG_WR;
    td_clk3 = SIG_WR | SIG_RD;
  } else {
    td_clk1 = 0;
    td_clk2 = SIG_WR;
    td_clk3 = 0;
  }

  if (SIG_CS) {
    td_clk1 |= SIG_CS;
    td_clk2 |= SIG_CS;
    td_clk3 |= SIG_CS;
  }

  if (iscmd) {
    td_clk1 |= SIG_A0;
    td_clk2 |= SIG_A0;
    td_clk3 |= SIG_A0;
  }
  
  for (i = 0; i < 8; i++)
    if (item & (1 << i))
       item_split |= dd->sdcd->signals[i];


  td_clk2 |= item_split;

  SDCONN_write(dd->sdcd, td_clk1, dd->sdcd->io_flags_writecmd);
  sdtools_nsleep(dd->delay);
  SDCONN_write(dd->sdcd, td_clk2, dd->sdcd->io_flags_writecmd | dd->sdcd->io_flags_writedata);
  sdtools_nsleep(dd->delay);
  SDCONN_write(dd->sdcd, td_clk3, dd->sdcd->io_flags_writecmd);
  sdtools_nsleep(dd->delay);
}



void serdisp_sed133x_writecmd(serdisp_t* dd, byte cmd) {
  serdisp_sed133x_transfer(dd, 1, cmd);
}


void serdisp_sed133x_writedata(serdisp_t* dd, byte data) {
  serdisp_sed133x_transfer(dd, 0, data);
}


/* callback-function for setting non-standard options */
void* serdisp_sed133x_getvalueptr (serdisp_t* dd, const char* optionname, int* typesize) {
  if (serdisp_compareoptionnames(dd, optionname, "OSCILLATOR")) {
    *typesize = sizeof(int);
    return &(serdisp_sed133x_internal_getStruct(dd)->fosc);
  } else if (serdisp_compareoptionnames(dd, optionname, "INTERFACE")) {
    *typesize = sizeof(int);
    return &(serdisp_sed133x_internal_getStruct(dd)->interfacemode);
  }
  return 0;
}

/* main functions */


/* *********************************
   serdisp_t* serdisp_sed133x_setup(sdcd, dispname, optionstring)
   *********************************
   sets up a display descriptor fitting to dispname and extra
   *********************************
   sdcd             ... output device handle (not used in here)
   dispname         ... display name (case-insensitive)
   optionstring     ... option string containing individual options
   *********************************
   returns a display descriptor
*/
serdisp_t* serdisp_sed133x_setup(const serdisp_CONN_t* sdcd, const char* dispname, const char* optionstring) {
  serdisp_t* dd;
  int scrbuf_columns;

  if (! (dd = (serdisp_t*)sdtools_malloc(sizeof(serdisp_t)) ) ) {
    sd_error(SERDISP_EMALLOC, "serdisp_sed133x_setup(): cannot allocate display descriptor");
    return (serdisp_t*)0;
  }
  memset(dd, 0, sizeof(serdisp_t));

  if (! (dd->specific_data = (void*) sdtools_malloc( sizeof(serdisp_sed133x_specific_t)) )) {
    free(dd);
    return (serdisp_t*)0;
  }

  memset(dd->specific_data, 0, sizeof(serdisp_sed133x_specific_t));

  /* "SED133X"-based displays supported in here */
  /* assign dd->dsp_id */
  if (serdisp_comparedispnames("SED133X", dispname))
    dd->dsp_id = DISPID_SED133X;
  else {  /* should not occur */
    sd_error(SERDISP_ENOTSUP, "display '%s' not supported by serdisp_specific_sed133x.c", dispname);
    return (serdisp_t*)0;
  }

  /* per display settings */

  dd->width             = 240;
  dd->height            = 64;
  dd->depth             = 1;
  dd->feature_contrast  = 0;
  dd->feature_backlight = 0;
  dd->feature_invert    = 0;
#if (SERDISP_VERSION_CODE > SERDISP_VERSION(1,97))
  dd->colour_spaces     = SD_CS_SCRBUFCUSTOM | SD_CS_GREYSCALE;
#endif

  /* max. delta for optimised update algorithm */
  dd->optalgo_maxdelta  = 3;

  dd->delay = 0;
  serdisp_sed133x_internal_getStruct(dd)->interfacemode = INTERFACE_8080;
  serdisp_sed133x_internal_getStruct(dd)->fosc = 9600;

  /* finally set some non display specific defaults */

  dd->curr_rotate       = 0;         /* unrotated display */
  dd->curr_invert       = 0;         /* display not inverted */

  /* supported output devices */
  dd->connection_types  = SERDISPCONNTYPE_PARPORT;

  dd->fp_init           = &serdisp_sed133x_init;
  dd->fp_update         = &serdisp_sed133x_update;
  dd->fp_close          = &serdisp_sed133x_close;
  dd->fp_setoption      = &serdisp_sed133x_setoption;
  dd->fp_setpixel       = &serdisp_sed133x_setpixel;
  dd->fp_getpixel       = &serdisp_sed133x_getpixel;
  dd->fp_getvalueptr    = &serdisp_sed133x_getvalueptr;

  serdisp_setupstructinfos(dd, serdisp_sed133x_wiresignals, serdisp_sed133x_wiredefs, serdisp_sed133x_options);

  /* parse and set options */
  if (serdisp_setupoptions(dd, dispname, optionstring) ) {
    free(dd->specific_data);
    free(dd);
    dd = 0;
    return (serdisp_t*)0;    
  }

  scrbuf_columns = (dd->width + dd->xcolgaps + 7) / 8; 

  /* calculate (non default) screen buffer size */
  dd->scrbuf_size = sizeof(byte) * (scrbuf_columns * ( dd->height +dd->ycolgaps )); 

  /* one byte is able to store 8 change infos */
  dd->scrbuf_chg_size = sizeof(byte) * ( ((int)((scrbuf_columns + 7)/8)) * ( dd->height +dd->ycolgaps ));

  return dd;
}



/* *********************************
   void serdisp_sed133x_init(dd)
   *********************************
   initialise a sed133x-based display
   *********************************
   dd     ... display descriptor
*/
void serdisp_sed133x_init(serdisp_t* dd) {
  int gfx_area, opt_vc, opt_cr, opt_tcr, opt_ap, opt_ws;

  opt_vc = (dd->width + dd->xcolgaps) / OPT_FX; /* characters per line */
  opt_cr = ((OPT_FX + 7) / 8) * opt_vc;  /* [CR] = Round([FX] / 8) * [VC] */
  
  opt_ws = (dd->height < 128) ? 0 : 1;

  gfx_area = opt_cr * dd->height;

  /* TCR must satisfy two conditions:

     1)  fosc >= ([TCR] * 9 + 1) * [LF] *fFR    LF = lines per frame => height; fFR = 70 Hz
         =>
         [TCR] = ((fosc / (height * 70)) - 1 ) / 9

     2)  [TCR] >= [CR] + 4

  */
  opt_tcr = ((serdisp_sed133x_internal_getStruct(dd)->fosc * 1000 / (dd->height * 70)) - 1) / 9;
  if (opt_tcr < opt_cr + 4) {
    opt_tcr = opt_cr + 4;
    sd_debug(1, "serdisp_sed133x_init(): fosc too low. corrected to meet condition [TRC] >= [CR] + 4");
  }

  opt_ap = opt_cr;   /* AP .. horizontal address range of virtual screen. == cr in our case */  

  sd_debug(2, "serdisp_sed133x_init(): gfx_area: %d, VC: %d, CR: %d, TCR: %d, AP: %d", gfx_area, opt_vc, opt_cr, opt_tcr, opt_ap);

  serdisp_sed133x_writecmd (dd, CMD_SYSTEMSET);
  serdisp_sed133x_writedata(dd, (OPT_DR<<7)+(OPT_TL<<6)+(OPT_IV<<5)+(1<<4)+(opt_ws<<3)+(OPT_M2<<2)+(OPT_M1<<1)+OPT_M0);
  serdisp_sed133x_writedata(dd, (OPT_WF<<7)+(OPT_FX-1));
  serdisp_sed133x_writedata(dd, (OPT_FY-1));                         /* [FR] - 1 */
  serdisp_sed133x_writedata(dd, (opt_cr-1));                         /* [CR] - 1 */
  serdisp_sed133x_writedata(dd, (opt_tcr-1));                        /* [TCR] - 1 */
  serdisp_sed133x_writedata(dd, (dd->height - 1));                   /* [LF] - 1     LF  lines per frame */
  serdisp_sed133x_writedata(dd, (byte)((opt_ap & 0x00FF)));          /* AP low byte */
  serdisp_sed133x_writedata(dd, (byte)( (opt_ap & 0xFF00) >> 8 ) );  /* AP high byte */

  serdisp_sed133x_writecmd (dd, CMD_SCROLL);
  serdisp_sed133x_writedata(dd, (byte)(OPT_SAD & 0x00FF));
  serdisp_sed133x_writedata(dd, (byte)((OPT_SAD & 0xFF00) >> 8));
  serdisp_sed133x_writedata(dd, ((opt_ws) ? (dd->height >> 1) : dd->height));
  serdisp_sed133x_writedata(dd, (byte)((OPT_SAD + gfx_area) & 0x00FF));
  serdisp_sed133x_writedata(dd, (byte)(((OPT_SAD + gfx_area) & 0xFF00) >> 8));
  serdisp_sed133x_writedata(dd, ((opt_ws) ? (dd->height >> 1) : dd->height));
  
  if (opt_ws) {
    serdisp_sed133x_writedata(dd, (byte)(OPT_SAD & 0x00FF));
    serdisp_sed133x_writedata(dd, (byte)((OPT_SAD & 0xFF00) >> 8));
  }

  serdisp_sed133x_writecmd (dd, CMD_CSRDIRRIGHT);  /* set cursor movement direction to right */
  
  serdisp_sed133x_writecmd (dd, CMD_HDOTSCR);
  serdisp_sed133x_writedata(dd, 0);   /* no horizontal scrolling */

  serdisp_sed133x_writecmd (dd, CMD_OVLAY);
  serdisp_sed133x_writedata(dd, (OPT_OV<<4)+(OPT_DM2<<3)+(OPT_DM1<<2)+(OPT_MX1<<1)+OPT_MX0);

  serdisp_sed133x_writecmd (dd, CMD_DISPON);
  if (opt_ws) {
    serdisp_sed133x_writedata(dd, (OPT_FP<<4)+(OPT_FP<<2)+(OPT_FC1<<1)+OPT_FC0);
  } else { 
    serdisp_sed133x_writedata(dd, (OPT_FP<<2)+(OPT_FC1<<1)+OPT_FC0);
  } 


  sd_debug(2, "serdisp_sed133x_init(): done with initialising");
}



/* *********************************
   void serdisp_sed133x_update(dd)
   *********************************
   updates the display using display-buffer scrbuf+scrbuf_chg
   *********************************
   dd     ... display descriptor
   *********************************

   the display is redrawn using a time-saving algorithm:
   
     background knowledge: after writing a page-entry to the display,
     the x-address is increased automatically =>
     * try to utilize this auto-increasing
       (if a whole horizontal line needs to be redrawn: only 4x PutCtrl per page will be needed)

     * on the other hand try to avoid writing of unchanged data: 
       best case: no need to change any data in a certain page: 0x PutCtrl (set page) + 0x PutCtrl (set xpos) + 0x PutData)

*/
void serdisp_sed133x_update(serdisp_t* dd) {
  int i, col;
  int max_col = ((dd->width + dd->xcolgaps + 7) / 8);
  byte data;

#ifdef OPT_USEOLDUPDATEALGO
  /* unoptimised display update (slow. all pixels are redrawn) */

  serdisp_sed133x_writecmd (dd, CMD_CSRW);
  serdisp_sed133x_writedata(dd, 0x00);
  serdisp_sed133x_writedata(dd, 0x00);

  serdisp_sed133x_writecmd (dd, CMD_MWRITE);

  for (i = 0; i < dd->height; i++) {
    for (col = 0; col < max_col; col++) {
      data = dd->scrbuf [ max_col * i + col];

      /* if (dd->curr_invert && !(dd->feature_invert)) */
      if (dd->curr_invert)
        data = ~data;

      serdisp_sed133x_writedata(dd, data);
    }
  }


#else /* OPT_USEOLDUPDATEALGO */

  /* display is drawn using an optimising algorithm which tries to only send as few data as possible to the display */

  /* more detailed explanations of principle: 
     see serdisp_specific_optrex323.c / i2c.c / pcd8544.c / sed153x.c / sed1565.c 
  */


  int delta; 
  int i_delta = 0; /* i_delta - i: how many columns to transfer in one take */

  i = 0;

  while (i < dd->scrbuf_size) {
    int y_i, col_i, l;

    int y = i / max_col;
    col = i % max_col;

    /* first changed column-page */
    if ( dd->scrbuf_chg[((max_col + 7) / 8) * y + (col / 8)] & ( 1 << (col%8) )) {
      i_delta = i;

      delta = 0;
      while (i_delta < dd->scrbuf_size-delta-1 && delta < dd->optalgo_maxdelta) {
        y_i = (i_delta+delta+1) / max_col;
        col_i = (i_delta+delta+1) % max_col;

        if ( dd->scrbuf_chg[((max_col + 7) / 8) * y_i + (col_i / 8)] & ( 1 << (col_i%8) )) {
          i_delta += delta+1;
          delta = 0;
        } else {
          delta++;
        }
      }

      serdisp_sed133x_writecmd (dd, CMD_CSRW);
      serdisp_sed133x_writedata(dd, (byte)(i%256) );
      serdisp_sed133x_writedata(dd, (byte)(i/256) );


      serdisp_sed133x_writecmd (dd, CMD_MWRITE);

      for (l = i; l <= i_delta; l++) {
        y_i = l / max_col;
        col_i = l % max_col;

        data = dd->scrbuf [l];

        /* if (dd->curr_invert && !(dd->feature_invert)) */
        if (dd->curr_invert)
          data = ~data;

        serdisp_sed133x_writedata(dd, data);
        dd->scrbuf_chg[ ((max_col + 7) / 8) * y_i + (col_i / 8)] &=  (0xFF ^ (1 << (col_i % 8)));
      }

      i = i_delta+1;

    } else {
      i++;
    }
  } /* while i < scrbuf_size */
#endif /* OPT_USEOLDUPDATEALGO */

  /* add an extra NOP to avoid erraneous pixels when releasing parport */
  serdisp_sed133x_writecmd (dd, CMD_DISPON);
}



/* *********************************
   int serdisp_sed133x_setoption(dd, option, value)
   *********************************
   change a display option
   *********************************
   dd      ... display descriptor
   option  ... name of option to change
   value   ... value for option
*/
int serdisp_sed133x_setoption(serdisp_t* dd, const char* option, long value) {

  if (dd->feature_backlight && serdisp_compareoptionnames(dd, option, "BACKLIGHT")) {
    if (value < 2) 
      dd->curr_backlight = (int)value;
    else
      dd->curr_backlight = (dd->curr_backlight) ? 0 : 1;
  } else if (serdisp_compareoptionnames(dd, option, "OSCILLATOR")) {
    int idx = serdisp_getoptionindex(dd, "OSCILLATOR");
    
    if (idx != -1) {
      if ( (serdisp_sed133x_options[idx].minval == -1 || (value >= serdisp_sed133x_options[idx].minval)) &&
           (serdisp_sed133x_options[idx].maxval == -1 || (value <= serdisp_sed133x_options[idx].maxval)) ) {
        int opt_vc, opt_cr, opt_tcr, opt_ap, opt_ws;

        serdisp_sed133x_internal_getStruct(dd)->fosc = value;

        opt_vc = (dd->width + dd->xcolgaps) / OPT_FX; /* characters per line */
        opt_cr = ((OPT_FX + 7) / 8) * opt_vc;  /* [CR] = Round([FX] / 8) * [VC] */
   
        opt_ws = (dd->height < 128) ? 0 : 1;
    
        opt_tcr = ((serdisp_sed133x_internal_getStruct(dd)->fosc * 1000 / (dd->height * 70)) - 1) / 9;
        if (opt_tcr < opt_cr + 4) {
          opt_tcr = opt_cr + 4;
          sd_debug(1, "serdisp_sed133x_setoption(): fosc too low. corrected to meet condition [TRC] >= [CR] + 4");
        }

        opt_ap = opt_cr;   /* AP .. horizontal address range of virtual screen. cr in our case; == cr in our case */  

        sd_debug(2, "serdisp_sed133x_setoption(): VC: %d, CR: %d, TCR: %d, AP: %d", opt_vc, opt_cr, opt_tcr, opt_ap);

        serdisp_sed133x_writecmd (dd, CMD_SYSTEMSET);
        serdisp_sed133x_writedata(dd, (OPT_DR<<7)+(OPT_TL<<6)+(OPT_IV<<5)+(1<<4)+(opt_ws<<3)+(OPT_M2<<2)+(OPT_M1<<1)+OPT_M0);
        serdisp_sed133x_writedata(dd, (OPT_WF<<7)+(OPT_FX-1));
        serdisp_sed133x_writedata(dd, (OPT_FY-1));                         /* [FR] - 1 */
        serdisp_sed133x_writedata(dd, (opt_cr-1));                         /* [CR] - 1 */
        serdisp_sed133x_writedata(dd, (opt_tcr-1));                        /* [TCR] - 1 */
        serdisp_sed133x_writedata(dd, (dd->height - 1));                   /* [LF] - 1     LF  lines per frame */
        serdisp_sed133x_writedata(dd, (byte)((opt_ap & 0x00FF)));          /* AP low byte */
        serdisp_sed133x_writedata(dd, (byte)( (opt_ap & 0xFF00) >> 8 ) );  /* AP high byte */

      } else {
        sd_debug(1, "serdisp_sed133x_setoption(): fosc out of bounds");
      }
    }
  
  } else {
    /* option not found here: try generic one in calling serdisp_setoption(); */
    return 0;
  }
  return 1;
}



/* *********************************
   void serdisp_sed133x_close(dd)
   *********************************
   close (switch off) display
   *********************************
   dd     ... display descriptor
*/
void serdisp_sed133x_close(serdisp_t* dd) {
  serdisp_sed133x_writecmd (dd, CMD_DISPOFF);
}




/* *********************************
   void serdisp_sed133x_setpixel(dd, x, y, colour)
   *********************************
   changes a pixel into the display buffer
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   colour ... monochrome: 0: clear (white), <>0: set (black)
*/
void serdisp_sed133x_setpixel (serdisp_t* dd, int x, int y, long colour) {
  int x_i = 0, y_i = 0, col;
  byte value_orig, value, mask;
  int cols = (dd->width + dd->xcolgaps + 7) / 8;

  if (dd->curr_rotate <= 1) {
    if (x >= dd->width || y >= dd->height || x < 0 || y < 0)
      return;
  } else {
    if (x >= dd->height || y >= dd->width || x < 0 || y < 0)
      return;
  }

  switch (dd->curr_rotate) {
    case 0:  /* 0 degrees */
      x_i = dd->xreloctab[x];
      y_i = dd->yreloctab[y];
      break;
    case 1:  /* 180 degrees */
      x_i = dd->xreloctab[dd->width  - 1 - x];
      y_i = dd->yreloctab[dd->height - 1 - y];
      break;
    case 2:  /* 90 degrees */
      x_i = dd->xreloctab[y];
      y_i = dd->yreloctab[dd->height  - 1 - x];
      break;
    case 3:  /* 270 degrees */
      x_i = dd->xreloctab[dd->width - 1 - y];
      y_i = dd->yreloctab[x];
      break;
  }


  col = x_i / 8;

  value_orig = dd->scrbuf[ cols * y_i + col ];

  mask = 1 << (7-(x_i % 8));

  value = (colour) ? (value_orig | mask) : (value_orig & (0xFF ^ mask)); 

  if (value_orig != value) {
    dd->scrbuf[ cols * y_i + col ] = value;
    dd->scrbuf_chg[ ((cols + 7) / 8) * y_i + (col / 8)] |=  (1 << (col % 8));
  }
}



/* *********************************
   int serdisp_sed133x_getpixel(dd, x, y)
   *********************************
   changes a pixel in the display buffer
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   *********************************
   returns the colour at (x/y)
*/
long serdisp_sed133x_getpixel (serdisp_t* dd, int x, int y) {
  int x_i = 0, y_i = 0, col;
  byte value, mask;
  int cols = (dd->width + dd->xcolgaps + 7) / 8;

  if (dd->curr_rotate <= 1) {
    if (x >= dd->width || y >= dd->height || x < 0 || y < 0)
      return 0;
  } else {
    if (x >= dd->height || y >= dd->width || x < 0 || y < 0)
      return 0;
  }

  switch (dd->curr_rotate) {
    case 0:  /* 0 degrees */
      x_i = dd->xreloctab[x];
      y_i = dd->yreloctab[y];
      break;
    case 1:  /* 180 degrees */
      x_i = dd->xreloctab[dd->width  - 1 - x];
      y_i = dd->yreloctab[dd->height - 1 - y];
      break;
    case 2:  /* 90 degrees */
      x_i = dd->xreloctab[y];
      y_i = dd->yreloctab[dd->height  - 1 - x];
      break;
    case 3:  /* 270 degrees */
      x_i = dd->xreloctab[dd->width - 1 - y];
      y_i = dd->yreloctab[x];
      break;
  }

  col = x_i / 8;

  value = dd->scrbuf[ cols * y_i + col ];

  mask = 1 << (7-(x_i % 8));

  return ((value & mask) ? 1 : 0); 
}


